/*
	File name:	2dto2d_misc.c
	Language:	C
	Type:		Subroutines
	Author:		Gang Liu
	Date:		9/12/98
	Function:	Contains several low level subroutines for the
			geometric calculations needed in the forward and
			backward perspective projection of the a-b plane
			in the object frame.

			Included routines are:
				get_rot_mat()
				uv2ab()
				abc2uv()

*/
#include <math.h>

#include "2dto2d_misc.h"

void	get_rot_mat(float omega, float phi, float kappa, float r[3][3])
/*
 Arguments:
	omega	--->	the rotation angle in radians about the x axis;
	phi	--->	the rotation angle in radians about the y axis;
	kappa	--->	the rotation angle in radians about the z axis;
	r	<---	the 3x3 rotation matrix of the object frame relative
			to the camera frame;

 Actions:
	The orientation of the object frame relative to the camera frame is
	determined by first rotating the camera frame about its x axis for
	the amount of 'omega', the rotating the rotated frame about its y
	axis for 'phi', and finally rotating the twice rotated frame about
	its z axis for 'kappa'. The rotation matrix 'r' representing the
	final orientation of the object frame relative to the camera frame
	is calculated.

 Return values:
	None.
*/
{
	double	c_omega, s_omega, c_phi, s_phi, c_kappa, s_kappa;

	c_omega = cos(omega);	s_omega = sin(omega);
	c_phi = cos(phi);	s_phi = sin(phi);
	c_kappa = cos(kappa);	s_kappa = sin(kappa);

	r[0][0]= c_phi*c_kappa;
	r[0][1]= s_omega*s_phi*c_kappa + c_omega*s_kappa;
	r[0][2]= -c_omega*s_phi*c_kappa + s_omega*s_kappa;
	r[1][0]= -c_phi*s_kappa;
	r[1][1]= -s_omega*s_phi*s_kappa + c_omega*c_kappa;
	r[1][2]= c_omega*s_phi*s_kappa + s_omega*c_kappa;
	r[2][0]= s_phi;
	r[2][1]= -s_omega*c_phi;
	r[2][2]= c_omega*c_phi;
}

void	uv2ab(float r[3][3], float x0, float y0, float z0, float f,
		float u, float v, float *a, float *b)
/*
 Arguments:
	r	--->	the 3x3 rotation matrix of the object frame relative
			to the camera frame;
	x0	--->	the x coordinate of the origin of the object frame
			in the camera frame;
	y0	--->	similar to x0, but for the y coordinate;
	z0	--->	similar to x0, but for the z coordinate;
	f	--->	the location of the imaging plane on the z axis;
	u	--->	the u coordinate of a point on the imaging plane;
	v	--->	the v coordinate of a point on the imaging plane;
	a	<---	the a coordinate of the point on the a-b plane of
			the object frame whose perspective projection on the
			imaging plane is (u,v)
	b	<---	similar to a, but for the b coordinate;

 Actions:
	The pose (location and orientation) of the object frame relative to
	the camera frame is specified by 'r', 'x0', 'y0', 'z0'; a point
	(u,v) on the imaging plane z=f is given; (u,v) is back projected to
	the plane c=0 in the object frame; the coordinates of the back
	projected point is calculated and returned as (a,b).

 Return values:
	None.
*/
{
	double	dx, dy, dz, factor;

	factor=1.0/((double)u*r[2][0]+(double)v*r[2][1]+(double)f*r[2][2]);

	dx=factor*(r[2][1]*((double)u*y0-(double)v*x0) +
			r[2][2]*((double)u*z0-(double)f*x0));
	dy=factor*(r[2][0]*((double)v*x0-(double)u*y0) +
			r[2][2]*((double)v*z0-(double)f*y0));
	dz=factor*(r[2][0]*((double)f*x0-(double)u*z0) +
			r[2][1]*((double)f*y0-(double)v*z0));

	*a=r[0][0]*dx + r[0][1]*dy + r[0][2]*dz;
	*b=r[1][0]*dx + r[1][1]*dy + r[1][2]*dz;
}

void	abc2uv(float r[3][3], float x0, float y0, float z0, float f,
		float a, float b, float c, float *u, float *v)
/*
 Arguments:
	r	--->	the 3x3 rotation matrix of the object frame relative
			to the camera frame;
	x0	--->	the x coordinate of the origin of the object frame
			in the camera frame;
	y0	--->	similar to x0, but for the y coordinate;
	z0	--->	similar to x0, but for the z coordinate;
	f	--->	the location of the imaging plane on the z axis;
	a	--->	the a coordinate of the point in the object frame;
	b	--->	similar to a, but for the b coordinate;
	c	--->	similar to a, but for the c coordinate;
	u	<---	the u coordinate of the projection of (a,b,c) on
			the imaging plane;
	v	<---	similar to u, but for the v coordinate;

 Actions:
	The pose (location and orientation) of the object frame relative to
	the camera frame is specified by 'r', 'x0', 'y0', 'z0'; a point
	(a,b,c) in the object frame is given; its projection onto the imaging
	plane z=f is calculated and returned as (u,v).

	Note: If (a,b,c) is such that it falls on the plane z=0 of the
	camera frame, it can never be projected to the imaging plane. In
	this case, the returned (u,v) value is unpredicted.

 Return values:
	None.
*/
{
	double	x, y, z, tmp;

	x=r[0][0]*a + r[1][0]*b + r[2][0]*c + x0;
	y=r[0][1]*a + r[1][1]*b + r[2][1]*c + y0;
	z=r[0][2]*a + r[1][2]*b + r[2][2]*c + z0;

	tmp=f/z;
	*u=tmp*x;
	*v=tmp*y;
}

