/*
	File name:	connect.c
	Language:	C
	Type:		Subroutines
	Author:		Gang Liu
	Date:		6/5/97
	Function:	Implementation of the Lumia, Shapiro, and Zuniga
			algorithm for the connected component labeling,
			according to Chapter 2.3.5 of the book
			"Computer and Robot Vision" by Haralick and Shapiro.

	Update:		#1
	Author:		Gang Liu
	Date:		12/21/97
	Purpose:	1. added the subroutine component_sizes() and
			   remove_component();
			2. added documentation for connected_component();
			3. changed the final label assignment part of
			   connected_component() from using multiple image
			   scans to using a look-up table, which makes the
			   procedure much quicker.

	Update:		#2
	Author:		Gang Liu
	Date:		12/27/97
	Purpose:	Previously the list of equivalent labels was
			organized improperly in that each equivalent pair
			does not always has one of the labels as the
			smallest label of their equivalent class. Problems
			were detected when the second (bottom-up) pass was
			performed from left to right. Although doing the
			second pass from right to left seemed to have fixed
			the problem, it is still not logically correct.
			The algorithm should work for the second pass to be
			done from left to right, as stated in the book.

			Main change, although not very much modification in
			the code, was made to update_equiv_list() such that
			after adding an entry to the equivalent label list,
			each node in the list has its first label as the
			smallest label (the equivalent label referred to in
			the book) for its equivalent class. Now it works on
			my test image, which gave problem before, whether
			SECOND_PASS_RIGHT_TO_LEFT is set to 1 or 0.
			However, setting SECOND_PASS_RIGHT_TO_LEFT to 1
			should make the code more efficient in execution.

			In doing the above, reorganized the code a little
			bit and added two low level routines
			add_node_to_list() and smallest_eq_label().

	Update:		#3
	Author:		Gang Liu
	Date:		7/20/98
	Purpose:	The kernel of the algorithm, i.e., the part for
			deciding which equivalent labels belong to the same
			class, was not implemented correctly before.
			Previous implementation is correct provided that
			each equivalent class has at most 2 members, or, the
			smallest member of each equivalent class appear in an
			equivalent pair with every other member of the
			class. This does not hold in real applications.

			update_equiv_list() is renamed to
			update_equiv_pairs(), and is modified so that it just
			produce a list of equivalent pairs, and the ordering
			of the two members in the pair is not important.

			Added two routines, find_equiv_classes() and
			mark_same_class(), to assign the smallest label in
			each equivalent class to all members of that class.
			It should be noted that mark_same_class() is a
			recursive procedure that performs a depth first
			search (DFS) grouping the labels into equivalent
			classes, as originally prescribed in the book.
			Now there is no difference if the second pass, the
			bottom-up pass, is done from left to right or from
			right to left. Therefore, the flag
			SECOND_PASS_RIGHT_TO_LEFT and its associated code
			are removed.

			Thanks to Yalin Wang (ylwang@george.ee.washington.edu)
			for telling me what data structure to use for the
			DFS procedure.

			Added free_list(), and removed smallest_eq_label().

*/
#include <malloc.h>
#include <stdio.h>	/* for NULL */

#include "connect.h"

typedef	struct EquivNode	Equiv;
struct EquivNode {
	int	s1, s2;
	Equiv	*next;
};

#ifndef	mymin
#define	mymin(x,y)	((x)<(y)?(x):(y))
#endif
#ifndef	mymax
#define	mymax(x,y)	((x)>(y)?(x):(y))
#endif

int	connected_component(char map[], int cols, int rows, int neighbors,
			int labelmap[], int *num_label);
int	component_sizes(int labelmap[], int cols, int rows,
			int num_label, long sizes[]);
int	remove_component(int labelmap[], int cols, int rows, int *num_label,
			int smallest);

static	int	new_label(int init);
static	void	get_neighbor_labels(int labelmap[], int cols, int rows,
			int x, int y, int neighbors, int labels[]);
static	int	min_neighbor_label(int labelmap[], int cols, int rows,
			int neighbors, int x, int y);

static	int	update_equiv_pairs(int labelmap[], int cols, int rows,
			int x, int y, int neighbors, Equiv *listhead, int *num);
static	int	find_equiv_classes(Equiv *pairs_head, Equiv *classes_head);
static void	mark_same_class(Equiv members[], int index, int label);

static	int	already_in_list(Equiv *listhead, Equiv *node);
static	int	add_node_to_list(Equiv *listhead, Equiv *node);
static	void	free_list(Equiv *listhead);

int	connected_component(char map[], int cols, int rows, int neighbors,
			int labelmap[], int *num_label)
/*
 Parameters:
	map	--->	the foreground/background bitmap image; represented
			by a 1D array of 'rows'*'cols' elements in the
			row-major order; 0 marks the background pixels
			and all non-zero values mark the foreground pixels;
	cols	--->	number of columns of the bitmap image;
	rows	--->	number of rows of the bitmap image;
	neighbors
		--->	the definition of connectivity; either 8 or 4;
			and is set to 4 it is neither 8 nor 4 on input;
	labelmap
		<---	array of labels;
			has 'cols'*'rows' elements, organized row-wise;
			each element takes value in [0,'num_label'], where 0
			is the label for the background locations, and
			consecutive positive integer values starting from 1
			are the labels for the connected components;
	num_label
		<---	total number of foreground connected components
			contained in the 'labelmap';

 Action:
	This is an implementation of the Lumia, Shapiro, and Zuniga
	algorithm for the connected component labeling, according to
	Chapter 2.3.5 of the book
		Haralick and Shapiro, "Computer and Robot Vision,"
		Addison-Wesley Publishing Company, Inc., 1992.

	Non-zero pixels in 'map' are taken as the foreground pixels; a
	connected component analysis with the connectivity specified by
	'neighbors' is applied to 'map'; as a result, a connected component
	labeling of the bitmap image 'map' is obtained in 'labelmap' and the
	number of components are obtained in 'num_label';

 Return values:
	0	success
	<0	memory allcation for some working buffers failed
*/
{
	int	i, j, k;
	long	ptr;

	int	thislabel, templabel;
	int	maxlabel, count, *lut;
	char	*inuse;

	int	num_eq_pairs, num_eq_classes;
	Equiv	equiv_pairs, equiv_classes, *p;

	if(neighbors!=8 && neighbors!=4) neighbors=4;

	/* initialize the label counter */
	new_label(1);

	for(ptr=0;ptr<(long)cols*rows;ptr++) labelmap[ptr]=0;

	/* top-down pass */
	for(i=0;i<rows;i++) {
		num_eq_classes=num_eq_pairs=0;
		equiv_classes.next=equiv_pairs.next=NULL;

		/* start processing the line */
		ptr=(long)i*cols;
		for(j=0;j<cols;j++) {
			if(!map[ptr++]) continue;

			templabel=min_neighbor_label(labelmap, cols, rows,
							neighbors, j, i);
			thislabel=(templabel ? templabel : new_label(0));
			labelmap[ptr-1]=thislabel;
			k=update_equiv_pairs(labelmap, cols, rows, j, i,
					neighbors, &equiv_pairs, &num_eq_pairs);
			if(k) {
				free_list(&equiv_pairs);
				return -1;
			}
		}

		k=find_equiv_classes(&equiv_pairs, &equiv_classes);
		if(k) {
			free_list(&equiv_pairs);
			if(equiv_classes.next!=NULL) free(equiv_classes.next);
			return -2;
		}

		/* re-assign the labels with their equivalent label */
		ptr=(long)i*cols;
		for(j=0;j<cols;j++) {
			thislabel=labelmap[ptr++];
			if(!thislabel) continue;
			p=equiv_classes.next;
			while(p!=NULL) {
				if(p->s1==thislabel) break;
				p=p->next;
			}
			if(p==NULL) continue;
			labelmap[ptr-1]=p->s2;
		}

		/* free up the memory of the equivalence link lists */
		free_list(&equiv_pairs);
		if(equiv_classes.next!=NULL) free(equiv_classes.next);
	}

	/* bottom-up pass */
	for(i=rows-1;i>=0;i--) {
		num_eq_classes=num_eq_pairs=0;
		equiv_classes.next=equiv_pairs.next=NULL;

		/* start processing the line */
		ptr=(long)i*cols;
		for(j=0;j<cols;j++) {
			if(!labelmap[ptr++]) continue;
			k=update_equiv_pairs(labelmap, cols, rows, j, i,
					neighbors, &equiv_pairs, &num_eq_pairs);
			if(k) {
				free_list(&equiv_pairs);
				return -3;
			}
		}

		k=find_equiv_classes(&equiv_pairs, &equiv_classes);
		if(k) {
			free_list(&equiv_pairs);
			if(equiv_classes.next!=NULL) free(equiv_classes.next);
			return -4;
		}

		/* re-assign the labels with their equivalent label */
		ptr=(long)i*cols;
		for(j=0;j<cols;j++) {
			thislabel=labelmap[ptr++];
			if(!thislabel) continue;
			p=equiv_classes.next;
			while(p!=NULL) {
				if(p->s1==thislabel) break;
				p=p->next;
			}
			if(p==NULL) continue;
			labelmap[ptr-1]=p->s2;
		}

		/* free up the memory of the equivalence link lists */
		free_list(&equiv_pairs);
		if(equiv_classes.next!=NULL) free(equiv_classes.next);
	}

	/* counting the number of distinct labels, and relabel them from 1 */
	*num_label=0;

	maxlabel=0;
	for(ptr=0;ptr<(long)rows*cols;ptr++) {
		thislabel=labelmap[ptr];
		if(!thislabel) continue;
		if(maxlabel<thislabel) maxlabel=thislabel;
	}

	inuse=(char*)malloc((maxlabel+1)*sizeof(char));
	lut=(int*)malloc((maxlabel+1)*sizeof(int));
	if(lut==NULL || inuse==NULL) return -4;

	/* see which of the labels in [0,maxlabel] are actually being used */
	for(i=0;i<=maxlabel;i++) inuse[i]=0;
	for(ptr=0;ptr<(long)rows*cols;ptr++) {
		thislabel=labelmap[ptr];
		if(!thislabel) continue;
		inuse[thislabel]=1;
	}
	inuse[0]=1;	/* background always in use */

	lut[0]=0;	/* background remain background */
	for(i=1,count=0;i<=maxlabel;i++)
		lut[i] = inuse[i] ? ++count : 0;

	for(ptr=0;ptr<(long)cols*rows;ptr++)
		labelmap[ptr]=lut[labelmap[ptr]];
	*num_label=count;

	free(inuse); free(lut);

	return 0;
}

static	int	min_neighbor_label(int labelmap[], int cols, int rows,
			int neighbors, int x, int y)
{
	int	k;
	int	minlabel;
static	int	labels[8];

	get_neighbor_labels(labelmap, cols, rows, x, y, neighbors, labels);

	/* first find a non-zero label */
	minlabel=0;
	for(k=0;k<neighbors;k++) {
		if(labels[k]) {
			minlabel=labels[k];
			break;
		}
	}
	if(minlabel==0)
		/* neighbors either not labeled or all background pixels */
		return minlabel;

	/* then find the minimum label */
	for(k=0;k<neighbors;k++) {
		if(labels[k]==0) continue;
		if(minlabel>labels[k]) minlabel=labels[k];
	}
	return minlabel;
}

static	int	update_equiv_pairs(int labelmap[], int cols, int rows,
			int x, int y, int neighbors, Equiv *listhead, int *num)
{
	int	k;
	int	thislabel;
static	int	labels[8];
	Equiv	node;

	thislabel=labelmap[(long)y*cols+x];
	get_neighbor_labels(labelmap, cols, rows, x, y, neighbors, labels);

	/* add a possible new entry to the equivalent list */
	for(k=0;k<neighbors;k++) {
		if( ! labels[k] ) continue;
		if( thislabel != labels[k] ) {
			node.s1 = mymin(thislabel, labels[k]);
			node.s2 = mymax(thislabel, labels[k]);
			if(!already_in_list(listhead, &node)) {
				if(add_node_to_list(listhead, &node)) return -1;
				(*num)++;
			}
		}
	}

	return 0;
}

static	int	add_node_to_list(Equiv *listhead, Equiv *node)
{
	Equiv	*p;

	p=(Equiv*)malloc(sizeof(Equiv));
	if(p==NULL) return -1;
	p->next = listhead->next;
	listhead->next = p;
	p->s1 = node->s1;
	p->s2 = node->s2;
	return 0;
}

static	void	free_list(Equiv *listhead)
{
	Equiv	*p;

	while(listhead->next!=NULL) {
		p=listhead->next;
		listhead->next=p->next;
		free(p);
	}
}

static	int	already_in_list(Equiv *listhead, Equiv *node)
{
	Equiv	*p;
	int	found;

	found=0;
	p=listhead->next;
	while(p!=NULL) {
		if(p->s1==node->s1 && p->s2==node->s2) {
			found=1;
			break;
		}
		p=p->next;
	}
	return found;
}

static	void	get_neighbor_labels(int labelmap[], int cols, int rows,
			int x, int y, int neighbors, int labels[])
{
static	int	neighbors_loc[8][2]={	{0,-1},{-1,0},{1,0},{0,1},
					{-1,-1},{1,-1},{-1,1},{1,1}};
	int	i, j, k;

	if((neighbors!=8) && (neighbors!=4)) neighbors=4;
	for(k=0;k<neighbors;k++) labels[k]=0;

	for(k=0;k<neighbors;k++) {
		i = y + neighbors_loc[k][1];
		j = x + neighbors_loc[k][0];
		if(i<0 || i>=rows || j<0 || j>=cols)
			labels[k]=0;
		else
			labels[k]=labelmap[(long)i*cols+j];
	}
	return;
}

static	int	new_label(int init)
{
static	int	counter;

	if(init) counter=-1;
	return ++counter;
}

int	component_sizes(int labelmap[], int cols, int rows,
			int num_label, long sizes[])
/*
 Parameters:
	labelmap
		--->	array of labels;
			has 'cols'*'rows' elements, organized row-wise;
			each element takes value in [0,'num_label'], where 0
			is the label for the background locations, and
			consecutive positive integer values start from 1 are
			the labels for the connected components;
			the output of connected_component() satisfies the
			requirements;
	cols	--->	number of columns of the underlying image;
	rows	--->	number of rows of the underlying image;
	num_label
		--->	total number of foreground connected components
			contained in the 'labelmap';
	sizes	<---	array of the sizes (number of pixels) of the
			foreground connected components in the 'labelmap';
			has 'num_label' elements, and memory buffer should
			have been allocated before calling;
			Notice: since the background is of no intereset,
			the index follow the C convention of starting from 0;
			therefore, sizes[i] gives the size of component whose
			label is i+1;

 Action:
	The size of each of the connected components in 'labelmap' is counted
	and returned in 'areas'.

 Return values:
	0	success
	-1	some label in 'labelmap' has a value not in [0,'num_label'];
		either 'num_label' is not correct, or 'labelmap' does not
		satisfy requirements; this can never happen if both of them
		are given by connected_component();
*/
{
	long	i, label;

	for(i=0;i<num_label;i++) sizes[i]=0;
	for(i=0;i<(long)rows*cols;i++) {
		label=labelmap[i];
		if(!label) continue; /* background pixel */
		if(label<0 || label>num_label) return -1;
		label--;
		sizes[label]++;
	}
	return 0;
}

int	remove_component(int labelmap[], int cols, int rows, int *num_label,
			int smallest)
/*
 Parameters:
	labelmap
		<-->	array of labels;
			has 'cols'*'rows' elements, organized row-wise;
			each element takes value in [0,'num_label'], where 0
			is the label for the background locations, and
			consecutive positive integer values start from 1 are
			the labels for the connected components;
			the output of connected_component() satisfies the
			requirements;
	cols	--->	number of columns of the underlying image;
	rows	--->	number of rows of the underlying image;
	num_label
		<-->	total number of foreground connected components
			contained in the 'labelmap';
			NOTE: value used at both input and output;
	smallest
		--->	the size of the smallest connected components to be
			kept; components smaller than this is removed;

 Action:
	The size (number of pixels) of each of the connected components
	in 'labelmap' is counted by calling component_sizes(); the labels
	whose components are smaller than 'smallest' are identified, and the
	label for the components are changed to 0, namely, that of the
	background; the remaining components are then relabeled to make
	'labelmap' valid. (Validness requires label 0 used for background, and
	consecutive positive integers starting from 1 used for foreground
	connected components.)

 Return values:
	0	success
	-1	memory allcation for some working buffers failed
	-2	component_sizes() failed
*/
{
	int	i, count;
	long	*sizes, *lut;

	sizes=(long*)malloc((*num_label)*sizeof(long));
	lut=(long*)malloc((*num_label+1)*sizeof(long));
	if(sizes==NULL || lut==NULL) return -1;

	i=component_sizes(labelmap, cols, rows, *num_label, sizes);
	if(i) {
		free(sizes); free(lut); return -2;
	}

	/* relabel the foreground components */

	lut[0]=0;	/* background remain background */
	for(i=1,count=0;i<=*num_label;i++)
		lut[i] = (sizes[i-1]<smallest) ? 0 : ++count;
	*num_label=count;

	for(i=0;i<(long)cols*rows;i++)
		labelmap[i]=lut[labelmap[i]];

	free(sizes); free(lut);
	return 0;
}

static	int	find_equiv_classes(Equiv *pairs_head, Equiv *classes_head)
{
	int	i, j, new;
	int	num_pairs, unique, num_classes;

	Equiv	*members, *p, node;

	classes_head->next=NULL;

	num_pairs=0;
	for(p=pairs_head->next;p!=NULL;p=p->next) num_pairs++;

	if(num_pairs==0) return 0;

	members=malloc(2*num_pairs*sizeof(Equiv));
	if(members==NULL) return -1;


	/* find a list of the elements involved in the equivalent pairs */
	unique=0;
	for(p=pairs_head->next;p!=NULL;p=p->next) {

		new=1;
		for(j=0;j<unique;j++)
			if(p->s1==members[j].s1) {
				new=0;
				break;
			}
		if(new) members[unique++].s1=p->s1;

		new=1;
		for(j=0;j<unique;j++)
			if(p->s2==members[j].s1) {
				new=0;
				break;
			}
		if(new) members[unique++].s1=p->s2;
	}

	/* sort the members according to their 's1' field */
	for(i=0;i<unique-1;i++)
		for(j=i+1;j<unique;j++)
			if(members[i].s1 > members[j].s1) {
				node=members[i];
				members[i]=members[j];
				members[j]=node;
			}

	/* for each element in the list just constructed, build a link list
	   of the elements which form equivalent pairs with it */
	for(i=0;i<unique;i++) {
		members[i].next=NULL;
		for(p=pairs_head->next;p!=NULL;p=p->next) {
			node=*p;
			if(node.s1==members[i].s1 || node.s2==members[i].s1) {
				if(node.s1==members[i].s1) node.s1=node.s2;
				for(j=0;j<unique;j++)
					if(node.s1==members[j].s1) node.s2=j;

				if(!already_in_list(&(members[i]), &node)) {
					j=add_node_to_list(&(members[i]),&node);
					if(j){
						for(j=i;j>=0;j--)
							free_list(&(members[j]));
						free(members);
						return -1;
					}
				}
			}
		}
	}

	/* mark all elements as not having been visited */
	for(i=0;i<unique;i++) members[i].s2=-1;

	num_classes=0;
	for(i=0;i<unique;i++) {
		if(members[i].s2 != -1) continue;
		mark_same_class(members, i, members[i].s1);
		num_classes++;
	}

	for(j=0;j<unique;j++) free_list(&(members[j]));

	for(j=0;j<unique-1;j++) members[j].next=&(members[j+1]);
	members[unique-1].next=NULL;

	classes_head->next=members;
	return 0;
}

static void	mark_same_class(Equiv members[], int index, int label)
{
	Equiv	*p;

	if(members[index].s2 != -1) return;
	members[index].s2=label;
	for(p=members[index].next;p!=NULL;p=p->next)
		mark_same_class(members, p->s2, label);
}
