/*
	File name:	otsu.c
	Language:	C
	Type:		Subroutine
	Author:		Gang Liu
	Date:		06/12/96
	Function:	This is the implementation of Otsu's method for
			deciding the threshold value given a histogram.
			The implementation uses a recursive relationship
			to speed up the calculation.

	Update:		#1
	Author:		Gang Liu
	Date:		3/12/98
	Purpose:
			  . modified a little in the part for finding 'a'
			    and 'b'; also put them before memory allocation;
			  . modified the part for checking the emptiness of
			    the histogram by using 'a' and 'b';
			  . added documentation;
*/
#include <malloc.h>
#ifndef	NULL
#include <stdio.h>
#endif

int	otsu(long hist[], int bins, int *T)
/*
 Arguments:
	hist	--->	the array containing the histogram based on which the
			threshold value is to be decided; has 'bins' elements; 
			all elements must be non-negative;
	bins	--->	the number of bins used in constructing 'hist';
	T	<---	the index number (0 to 'bins'-1) of the bin whose
			representative data value is the optimal threshold;

 Actions:
	The histogram of some discrete data values, possibly quantized from
	continuous data, is taken as the input. Otsu's method for deciding the
	optimal threshold is carried out using the recursive relationship
	in the formula that speeds up the calculation.

	Uniform quantization of the data is assumed, or in other words, the
	bins used in obtaining the histogram are of the same size, as is the
	most common case in digital image processing. This routine is not
	interested in the absolute data values associated with the bins used
	in obtaining the histogram. Instead, it return the index of the bin
	whose corresponding representative absolute data value is the optimal
	value to threshold the original data.

	The histogram, which is the counts of data points falling in the bins,
	is non-negative by definition. Normally the data sample is non-empty
	and the histogram sums to at least 1. However, if the sum is 0 showing
	an empty data sample, the middle bin is returned as the expected
	threshold.

 Return values:
	0	success
	-1	invalid histogram; negative entry in 'hist';
	-2	failure in memory allocation for temporary memory buffers;
*/
{
	int	i, a, b;
	double	dtemp;
	double	sum, mu, _1_over_sum;
	double	*p, *q, *mu1, *mu2, *sigma;

	if(bins<=0) return 0;
	*T=bins/2;
	if(bins<=1) return 0;
	for(i=0;i<bins;i++)
		if(hist[i]<0) return -1;

	for(i=0;i<bins;i++) if(hist[i]) break;
	a=i;
	for(i=bins-1;i>=0;i--) if(hist[i]) break;
	b=i;

	if(b<a) return 0;

	p=(double*)malloc(bins*sizeof(double));
	q=(double*)malloc(bins*sizeof(double));
	mu1=(double*)malloc(bins*sizeof(double));
	mu2=(double*)malloc(bins*sizeof(double));
	sigma=(double*)malloc(bins*sizeof(double));
	if(p==NULL || q==NULL || mu1==NULL || mu2==NULL || sigma==NULL) {
		if(p!=NULL) free(p);
		if(q!=NULL) free(q);
		if(mu1!=NULL) free(mu1);
		if(mu2!=NULL) free(mu2);
		if(sigma!=NULL) free(sigma);
		return -2;
	}

	sum=0.0;
	for(i=0;i<bins;i++) {
		p[i]=hist[i];
		sum+=p[i];
	}
	_1_over_sum=1.0/sum;

	mu=0.0;
	for(i=a;i<=b;i++) {
		p[i] *= _1_over_sum;
		mu+=i*p[i];
	}

	q[a]=p[a];
	mu1[a]=a;
	mu2[a]=(mu - q[a]*mu1[a]) / (1. - q[a]);
	dtemp=mu1[a]-mu2[a];
	sigma[a]=q[a]*(1.-q[a])*dtemp*dtemp;

	for(i=a+1;i<=b-1;i++) {
		q[i] = q[i-1]+p[i];
		mu1[i] = (q[i-1]*mu1[i-1] + i*p[i]) / q[i];
		mu2[i] = (mu - q[i]*mu1[i]) / (1. - q[i]);
		dtemp = mu1[i]-mu2[i];
		sigma[i] = q[i] * (1-q[i]) * dtemp * dtemp;
	}

	dtemp=sigma[a];
	*T=a;
	for(i=a+1;i<=b-1;i++) {
		if(sigma[i]>dtemp) {
			dtemp=sigma[i];
			*T=i;
		}
	}

	free(p); free(q); free(mu1); free(mu2); free(sigma);
	return 0;
}

