/*
	File name:	zhmorph.c
	Type:		subroutines
	Author:		Gang Liu
	Date:		2/11/2000
	Function:	Subroutines for some basic zero-height gray-scale
			morphological operations.
*/
#include <stdio.h>	/* for NULL */
#include <malloc.h>
#include <memory.h>
#include <math.h>

#include "zhgmorph.h"

static	int	myint(float f)
{
	int	ff=floor(f);
	return ((f-ff >= 0.5) ? ff+1 : ff);
}

void	zhgopen(int **inimage, int rows, int cols, MSE *se, int **outimage)
{
	int	i;
	zhgerosion(inimage, rows, cols, se, outimage);
	for(i=0;i<rows;i++)
		memcpy(inimage[i], outimage[i], cols*sizeof(int));
	zhgdilation(inimage, rows, cols, se, outimage);
	return;
}

void	zhgclose(int **inimage, int rows, int cols, MSE *se, int **outimage)
{
	int	i;
	zhgdilation(inimage, rows, cols, se, outimage);
	for(i=0;i<rows;i++)
		memcpy(inimage[i], outimage[i], cols*sizeof(int));
	zhgerosion(inimage, rows, cols, se, outimage);
	return;
}

void	zhgdilation(int **inimage, int rows, int cols, MSE *se, int **outimage)
{
	int	i, j, k, x, y;
	int	firstround, maxval=0;

	for(i=0;i<rows;i++)
		for(j=0;j<cols;j++) {
			firstround=1;
			for(k=0;k<se->size;k++) {
				y = i + se->y[k];
				x = j + se->x[k];
				if(y<0 || y>=rows || x<0 || x>=cols) continue;
				if(inimage[y][x]>maxval || firstround) {
					maxval=inimage[y][x];
					firstround=0;
				}
			}
			/* if the SE fall entirely outside of image,
			   just copy the input */
			outimage[i][j]= (firstround ? inimage[i][j] : maxval);
		}
}

void	zhgerosion(int **inimage, int rows, int cols, MSE *se, int **outimage)
{
	int	i, j, k, x, y;
	int	firstround, minval=0;

	for(i=0;i<rows;i++)
		for(j=0;j<cols;j++) {
			firstround=1;
			for(k=0;k<se->size;k++) {
				y = i + se->y[k];
				x = j + se->x[k];
				if(y<0 || y>=rows || x<0 || x>=cols) continue;
				if(inimage[y][x]<minval || firstround) {
					minval=inimage[y][x];
					firstround=0;
				}
			}
			/* if the SE fall entirely outside of image,
			   just copy the input */
			outimage[i][j]= (firstround ? inimage[i][j] : minval);
		}
}

void	free_MSE(MSE *se)
{
	if( se->x != NULL ) free(se->x); se->x=NULL;
	if( se->y != NULL ) free(se->y); se->y=NULL;
	se->size=0;
	return;
}

int	block_MSE(int rows, int cols, int left, int top, MSE *se)
{
	int	i, j, k;

	se->size = k=rows*cols;
	se->x = (int*)malloc(k*sizeof(int));
	se->y = (int*)malloc(k*sizeof(int));
	if(se->x==NULL || se->y==NULL) {
		free_MSE(se);
		return -1;
	}

	k=0;
	for(i=0;i<rows;i++)
		for(j=0;j<cols;j++) {
			se->y[k] = i+top;
			se->x[k] = j+left;
			k++;
		}
	return 0;
}

int	disk_MSE(float R, int cen_x, int cen_y, MSE *se)
{
	int	i, j, k, rr;

	rr=R; if(R>rr) rr++;
	k=(2*rr+1)*(2*rr+1);
	se->x = (int*)malloc(k*sizeof(int));
	se->y = (int*)malloc(k*sizeof(int));
	if(se->x==NULL || se->y==NULL) {
		free_MSE(se);
		return -1;
	}

	k=0;
	for(i=-rr;i<=rr;i++)
		for(j=-rr;j<=rr;j++) {
			if(i*i+j*j>R*R) continue;
			se->y[k]=cen_y+i;
			se->x[k]=cen_x+j;
			k++;
		}
	se->size=k;

	return 0;
}

int	line_MSE(int x1, int y1, int x2, int y2, MSE *se)
{
	int	k, xstep, ystep;
	int	dx, dy;
	float	slope;

	xstep=ystep=1;
	dx=x2-x1; if(dx<0) { xstep=-1; dx=-dx;}
	dy=y2-y1; if(dy<0) { ystep=-1; dy=-dy;}

	se->size = k=(dx>=dy ? dx : dy)+1;
	se->x = (int*)malloc(k*sizeof(int));
	se->y = (int*)malloc(k*sizeof(int));
	if(se->x==NULL || se->y==NULL) {
		free_MSE(se);
		return -1;
	}

	if(dx>=dy) {
		/* sloped close to x-axis, quantize according to x-axis */
		slope=dy/dx;
		for(k=0;k<=dx;k++) {
			se->x[k]=x1+xstep*k;
			se->y[k]=myint(y1+ystep*k*slope);
		}
	}
	else {
		/* sloped close to y-axis, quantize according to y-axis */
		slope=dx/dy;
		for(k=0;k<=dy;k++) {
			se->y[k]=y1+ystep*k;
			se->x[k]=myint(x1+xstep*k*slope);
		}
	}

	return 0;
}
