/*
	File name:	zhgmorphpgm.c
	Type:		C main file.
	Author:		Gang Liu
	Date:		2/11/2000
	Function:	Perform some gray-scale morphological operations
			using zero-height structuring elements on an input
			gray-scale PGM image and save the result to an output
			gray-scale PGM image.

			Non-zero pixels in the input image are taken to
			be the fore-ground pixels. In the output image,
			fore-ground pixels are marked by 255 and back-ground
			pixels are marked by 0.
*/
#include <stdio.h>

#include "must.h"
#include "pgm.h"
#include "zhgmorph.h"

#define	DISK	0
#define	SQUARE	1
#define	HLINE	2
#define	VLINE	3
#define	NWLINE	4
#define	NELINE	5

void	usage(FILE *fp, char *cmd);

int	main(int argc, char *argv[])
{
	char	op;
	int	SEtype, SEsize, orig;
	char	*inname, *outname;

	int	xcen, ycen, R, left, top, x1, y1, x2, y2;
	MSE	se;

	uchar	**image;
	int	rows, cols;
	int	**inimg, **outimg;

	int	i, j;

	if(argc<7) {
		usage(stderr, argv[0]);
		return -1;
	}

	inname=argv[1];
	i =sscanf(argv[2], "%c", &op);
	i+=sscanf(argv[3], "%d", &SEtype);
	i+=sscanf(argv[4], "%d", &SEsize);
	i+=sscanf(argv[5], "%d", &orig);
	outname=argv[6];
	if(i!=4) {
		usage(stderr, argv[0]);
		return -1;
	}

	if(	(op!='d' && op!='e' && op!='o' && op!='c') ||
		(SEtype!=DISK && SEtype!=SQUARE &&
		 SEtype!=HLINE && SEtype!=VLINE &&
		 SEtype!=NWLINE && SEtype!=NELINE)	||
		(SEsize<=0 || SEsize%2!=1) ||
		(orig<0 || orig>4)
	  ) myerror(-2, stderr, "Invalid input parameter(s).\n");

	R=SEsize/2;
	xcen=ycen=left=top=x1=y1=x2=y2=0;
	switch(SEtype) {
		case DISK:
			switch(orig) {
				case 0: xcen=ycen=0; break;
				case 1: xcen=ycen=R; break;
				case 2: xcen=R; ycen=-R; break;
				case 3: xcen=-R; ycen=R; break;
				case 4: xcen=ycen=-R; break;
			}
			i=disk_MSE(SEsize/2, xcen, ycen, &se);
			break;
		case SQUARE:
			switch(orig) {
				case 0: left=-R; top=-R; break;
				case 1: left=top=0; break;
				case 2: left=0; top=-(SEsize-1); break;
				case 3: left=-(SEsize-1); top=0; break;
				case 4: left=top=-(SEsize-1); break;
			}
			i=block_MSE(SEsize, SEsize, left, top, &se);
			break;
		case HLINE:
			y1=y2=0;
			switch(orig) {
				case 0: x1=-(x2=R); break;
				case 1:
				case 2: x1=0; x2=SEsize-1; break;
				case 3:
				case 4: x1=-(SEsize-1); x2=0; break;
			}
			i=line_MSE(x1, y1, x2, y2, &se);
			break;
		case VLINE:
			x1=x2=0;
			switch(orig) {
				case 0: y1=-(y2=R); break;
				case 1:
				case 3: y1=0; y2=SEsize-1; break;
				case 2:
				case 4: y1=-(SEsize-1); y2=0; break;
			}
			i=line_MSE(x1, y1, x2, y2, &se);
			break;
		case NWLINE:
			switch(orig) {
				case 0: x1=y1=-R; x2=y2=R; break;
				case 1: x1=y1=0; x2=y2=SEsize-1; break;
				case 2: x1=y2=0; y1=-(x2=SEsize-1); break;
				case 3: x2=y1=0; x1=-(y2=SEsize-1); break;
				case 4: x2=y2=0; x1=y1=-(SEsize-1); break;
			}
			i=line_MSE(x1, y1, x2, y2, &se);
			break;
		case NELINE:
			switch(orig) {
				case 0: x1=y2=R; x2=y1=-R; break;
				case 1: x1=y2=SEsize-1; x2=y1=0; break;
				case 2: y1=-(x1=SEsize-1); x2=y2=0; break;
				case 3: x1=y1=0; x2=-(y2=SEsize-1); break;
				case 4: x1=y2=0; x2=-(y1=SEsize-1); break;
			}
			i=line_MSE(x1, y1, x2, y2, &se);
			break;
	}
	if(i) myerror(-3, stderr, "Error making the structuring element.\n");

	i=ReadPGM(inname, &cols, &rows, &image);
	if(i) myerror(-4, stderr, "Error reading input PGM image file.\n");

	inimg=(int**)alloc2d(cols, rows, sizeof(int));
	outimg=(int**)alloc2d(cols, rows, sizeof(int));
	if(inimg==NULL || outimg==NULL)
		myerror(-5, stderr, "Error allocating memory.\n");


	for(i=0;i<rows;i++)
		for(j=0;j<cols;j++) inimg[i][j]=image[i][j];

	switch(op) {
		case 'd':
			zhgdilation(inimg, rows, cols, &se, outimg);
			break;
		case 'e':
			zhgerosion(inimg, rows, cols, &se, outimg);
			break;
		case 'o':
			zhgopen(inimg, rows, cols, &se, outimg);
			break;
		case 'c':
			zhgclose(inimg, rows, cols, &se, outimg);
			break;
	}

	for(i=0;i<rows;i++)
		for(j=0;j<cols;j++) image[i][j]=outimg[i][j];

	i=WritePGM(outname, cols, rows, image);
	if(i) myerror(-5, stderr, "Error saving output file.\n");

	free2d((void**)image); free2d((void**)inimg); free2d((void**)outimg);

	return 0;
}

void	usage(FILE *fp, char *cmd)
{
fprintf(fp, "Function of %s:\n\
This utility takes an input PGM file as a gray-scale image and performs\n\
some zero-height gray-scale morphological operations on it. The result is\n\
a gray-scale image saved in a PGM file\n\n", cmd);
fprintf(fp,
"Usage:\n\t %s <PGM1> <op> <SE> <SE size> <orig> <PGM2>\n\
where\n\
<PGM1> (char*) is the name of the input image file;\n\
<op> (char) the morphological operation you want to perform:\n\
		'd' dilation; 'e' erosion; 'o' openning; 'c' closing.\n\
<SE> (int) the structuring element you want to use:\n\
		0 for circular disk; 1 for square;\n\
		2 for horizontal line; 3 for vertical line;\n\
		4 for 45 degree line (northwest to southeast);\n\
		5 for 135 degree line (northeast to southwest);\n\
<SE size> (int) size of the SE:\n\
		diameter (SE=0), side length (SE=1), length (SE=2 to 5);\n\
		should be an odd integer number;\n\
<orig> (int) location of the SE's origin;\n\
		0 for center;\n\
		1 for top-left corner; 2 bottom-left corner;\n\
		3 for top-right corner; 4 bottom-right corner;\n\
<PGM2> (char*) is the name of the output image file.\n\n", cmd);
}

